#include "ampapp.h"
#include "ampparser.h"
#include "constants.h"
#include "controllers/controllersstorage.h"
#include "mainwindow/mainwindow.h"

#include <iostream>
#include <QStandardItem>

class AMPAppPrivate
{
public:
    AMPAppPrivate(QApplication *app, DataSourceInterface *source, MainWindow *window)
        : m_app(app)
        , m_dataSource(source)
        , m_mainWindow(window)
        , m_interfaces({RPM_INTERFACE_NAME, APT_INTERFACE_NAME, REPO_INTERFACE_NAME})
        , m_model(new QStandardItemModel())
        , m_controllers(new ControllersStorage(source))
        , m_options(new CommandLineOptions())
        , m_parser(new AMPParser(*app))
        , m_parserResult{}
        , m_parseErrorMessage{}
        , m_interfaceDisplayNames{{RPM_INTERFACE_NAME, RPM_INTERFACE_DISPLAY_NAME},
                                  {APT_INTERFACE_NAME, APT_INTERFACE_DISPLAY_NAME},
                                  {REPO_INTERFACE_NAME, REPO_INTERFACE_DISPLAY_NAME}}
    {}

    QApplication *m_app{nullptr};

    DataSourceInterface *m_dataSource{nullptr};

    std::unique_ptr<MainWindow> m_mainWindow{nullptr};

    std::vector<QString> m_interfaces;

    std::unique_ptr<QStandardItemModel> m_model;

    std::unique_ptr<ControllersStorage> m_controllers;

    std::unique_ptr<CommandLineOptions> m_options;
    std::unique_ptr<AMPParser> m_parser;
    AMPParser::AMPParseResult m_parserResult;
    QString m_parseErrorMessage;

    QMap<QString, QString> m_interfaceDisplayNames;

private:
    AMPAppPrivate(const AMPAppPrivate &) = delete;
    AMPAppPrivate(AMPAppPrivate &&)      = delete;
    AMPAppPrivate &operator=(const AMPAppPrivate &) = delete;
    AMPAppPrivate &operator=(AMPAppPrivate &&) = delete;
};

AMPApp::AMPApp(QApplication *app, DataSourceInterface *source, QObject *parent)
    : QObject(parent)
    , d(new AMPAppPrivate(app, source, new MainWindow()))
{
    buildModel();

    d->m_mainWindow->setModel(d->m_model.get());

    connect(d->m_mainWindow.get(), &MainWindow::managerChanged, this, &AMPApp::onManagerChanged);
}

AMPApp::~AMPApp() {}

int AMPApp::runApp()
{
    d->m_parserResult = d->m_parser->parseCommandLine(d->m_options.get(), &d->m_parseErrorMessage);

    if (d->m_parserResult == AMPParser::AMPParseResult::CommandLineError)
    {
        std::cerr << d->m_parseErrorMessage.toStdString() << std::endl;
        d->m_parser->showHelp();
        return 1;
    }

    if (d->m_parserResult == AMPParser::AMPParseResult::CommandLineHelpRequested)
    {
        d->m_parser->showHelp();
        return 0;
    }

    if (d->m_parserResult == AMPParser::AMPParseResult::CommandLineVersionRequested)
    {
        d->m_parser->showVersion();
        return 0;
    }

    ControllerInterface *controller = nullptr;

    if (d->m_parserResult == AMPParser::AMPParseResult::CommandLineObjectRequested)
    {
        auto selectedObj = d->m_options->objectName;
        controller       = d->m_controllers->getControllerByPath(selectedObj);
    }

    if (!controller)
    {
        controller = d->m_controllers->getControllerByIface(DEFAULT_INTERFACE_NAME);
    }

    d->m_mainWindow->setController(controller);

    controller->setWindow(d->m_mainWindow.get());

    controller->selected();

    d->m_mainWindow->show();

    d->m_app->exec();

    return 0;
}

void AMPApp::buildModel()
{
    for (QString &iface : d->m_interfaces)
    {
        QStringList pathsList = d->m_dataSource->getObjectsPathsByIface(iface);
        if (pathsList.isEmpty())
        {
            continue;
        }
        QString name        = d->m_interfaceDisplayNames[iface];
        QStandardItem *item = new QStandardItem(name);
        item->setData(QVariant(iface), INTERFACE_ROLE);
        d->m_model->appendRow(item);
    }
}

void AMPApp::onManagerChanged(const QItemSelection &newSelection, const QItemSelection &previousSelection)
{
    if (newSelection.isEmpty())
    {
        return;
    }

    chooseController(LOAD_INTERFACE_NAME);

    QModelIndex currentIndex    = newSelection.indexes().at(0);
    QStandardItem *item         = static_cast<QStandardItem *>(currentIndex.internalPointer());
    QStandardItem *selectedItem = item->child(currentIndex.row());
    QString iface               = selectedItem->data(INTERFACE_ROLE).toString();
    chooseController(iface);
}

void AMPApp::chooseController(QString interface)
{
    ControllerInterface *controller = d->m_controllers->getControllerByIface(interface);

    if (!controller)
    {
        //TO DO send warning
        controller = d->m_controllers->getControllerByIface(DEFAULT_INTERFACE_NAME);
    }

    d->m_mainWindow->getController()->setWindow(nullptr);
    d->m_mainWindow->getController()->deSelected();

    d->m_mainWindow->setController(controller);
    controller->setWindow(d->m_mainWindow.get());

    controller->selected();

    d->m_app->processEvents();
}
