#include "aptcontroller.h"
#include "../datasource/dbusdatasource.h"
#include "aptbuttonswidget.h"
#include "aptcontrollerwidget.h"
#include "constants.h"
#include "package.h"

#include <set>
#include <QDate>
#include <QDebug>
#include <QInputDialog>
#include <QMessageBox>

class AptControllerPrivate
{
public:
    AptControllerPrivate()
        : forInstallation()
        , forRemoval()
        , m_buttonWidget(nullptr)
        , m_controllerWidget(nullptr)
        , m_modelData()
        , isFirstSelection(true)
        , m_selected(false)
    {}

    AptControllerPrivate(const AptControllerPrivate &) = delete;
    AptControllerPrivate(AptControllerPrivate &&)      = delete;

    AptControllerPrivate &operator=(const AptControllerPrivate &) = delete;
    AptControllerPrivate &operator=(AptControllerPrivate &&) = delete;

    QSet<QString> forInstallation;
    QSet<QString> forRemoval;

    AptButtonsWidget *m_buttonWidget;
    AptControllerWidget *m_controllerWidget;

    std::vector<Package> m_modelData;

    bool isFirstSelection;

    bool m_selected;
};

AptController::AptController(DataSourceInterface *dataSource)
    : BaseController(dataSource)
    , d(new AptControllerPrivate())
{
    QObject::connect(dynamic_cast<DBusDataSource *>(getDataSource()),
                     SIGNAL(aptUpdateFinished(QDBusPendingCallWatcher *)),
                     this,
                     SLOT(onAptUpdateFinished(QDBusPendingCallWatcher *)));
    updateModelData();
}

AptController::~AptController() {}

QString AptController::getInterface()
{
    return APT_INTERFACE_NAME;
}

QString AptController::getPath()
{
    return APT_OBJECT_PATH;
}

void AptController::selected()
{
    d->m_buttonWidget     = new AptButtonsWidget(this);
    d->m_controllerWidget = new AptControllerWidget(this);

    getWindow()->setButtonWidget(d->m_buttonWidget);
    getWindow()->setControllerWidget(d->m_controllerWidget);
    getWindow()->selectInterface(getInterface());

    d->m_controllerWidget->setPackagesData(d->m_modelData);

    d->m_controllerWidget->setFilter(QString());
    getWindow()->setSearchText(QString());

    d->m_selected = true;

    if (!d->isFirstSelection)
    {
        return;
    }

    d->isFirstSelection = false;

    QString lastUpdate;
    auto backendResult = getDataSource()->aptLastUpdate();
    if (backendResult.has_value())
    {
        lastUpdate = backendResult.value();
    }
    else
    {
        qWarning() << "Failed to get last package list update date from APT";
    }

    if (QDate::currentDate() == QDate::fromString(lastUpdate, "yyyy-MM-dd"))
    {
        return;
    }

    if (QMessageBox::Yes
        == QMessageBox::warning(d->m_controllerWidget,
                                tr("APT update is required."),
                                tr("It seems your package lists haven't been updated in a while. Would you "
                                   "like to update them?"),
                                QMessageBox::Yes | QMessageBox::No,
                                QMessageBox::Yes))
    {
        onUpdateButtonPressed();
    }
}

void AptController::deSelected()
{
    d->forInstallation.clear();
    d->forRemoval.clear();
    d->m_selected = false;
}

void AptController::onWindowClose() {}

void AptController::update()
{
    deSelected();
    updateModelData();
    selected();
}

void AptController::onFilterTextChanged(const QString text)
{
    d->m_controllerWidget->setFilter(text);
}

void AptController::onApplyButtonPressed()
{
    if (d->forInstallation.isEmpty() && d->forRemoval.isEmpty())
    {
        return;
    }

    auto acknowledgement = QMessageBox::question(
        d->m_controllerWidget,
        QObject::tr("Apply changes?"),
        QObject::tr("The following packages will be removed:\n")
            + QStringList(QList<QString>(d->forRemoval.begin(), d->forRemoval.end())).join(";\n")
            + QObject::tr("\nAnd The foolwing packages will be installed:\n")
            + QStringList(QList<QString>(d->forInstallation.begin(), d->forInstallation.end())).join(";\n"),
        QMessageBox::Apply | QMessageBox::Abort,
        QMessageBox::Abort);

    if (acknowledgement == QMessageBox::Apply)
    {
        QString pkgs("");
        for (auto pkg : d->forInstallation)
        {
            pkgs += QString("%1 ").arg(pkg);
        }
        for (auto pkg : d->forRemoval)
        {
            pkgs += QString("%1- ").arg(pkg);
        }
        getDataSource()->aptInstall(pkgs);
    }

    update();
}
void AptController::onUpdateButtonPressed()
{
    d->m_buttonWidget->setProgress(0);
    disableButtons();
    getDataSource()->aptUpdate();
    d->m_buttonWidget->setProgress(33);
}

void AptController::onAptUpdateFinished(QDBusPendingCallWatcher *call)
{
    // TODO: add an error message if call finished with non-zero exit code
    d->m_buttonWidget->setProgress(66);
    update();
    enableButtons();
    call->deleteLater();
    d->m_buttonWidget->setProgress(100);
    d->m_buttonWidget->hideProgress();
}

int AptController::onItemClicked(const Package &pkg)
{
    auto pkgName = pkg.m_name;

    if (d->forInstallation.remove(pkgName))
    {
        return 0;
    }
    if (d->forRemoval.remove(pkgName))
    {
        return 0;
    }

    if (pkg.m_installed)
    {
        d->forRemoval.insert(pkgName);
    }
    else
    {
        d->forInstallation.insert(pkgName);
    }

    return 1;
}

void AptController::disableButtons()
{
    d->m_buttonWidget->disableButtons();
}

void AptController::enableButtons()
{
    d->m_buttonWidget->enableButtons();
}

void AptController::updateModelData()
{
    d->m_modelData.clear();
    std::vector<QString> packages;
    QStringList installedPackages;

    auto listAllResult = getDataSource()->aptListAllPackages();
    if (listAllResult.has_value())
    {
        packages = listAllResult.value();
    }
    else
    {
        qWarning() << "Failed to get package list from APT";
    }

    auto listResult = getDataSource()->rpmList();
    if (listResult.has_value())
    {
        installedPackages = listResult.value();
    }
    else
    {
        qWarning() << "Failed to get installed packages list from RPM";
    }

    QRegularExpression regExp("^(.+)-((?:\\d+(?:\\.\\w+)*|\\w*)-alt\\d+.*).(x86_64|ppc64le|i586|armh|aarch64|noarch)$");
    for (auto &pkg : installedPackages)
    {
        auto match = regExp.match(pkg);
        if (match.hasMatch())
        {
            pkg = match.captured(1);
        }
        else
        {
            qWarning() << QString("Could not get package name of '%1'").arg(pkg);
        }
    }

    QSet<QString> installedSet(installedPackages.begin(), installedPackages.end());
    for (auto pkg : packages)
    {
        d->m_modelData.push_back(Package(pkg, installedSet.contains(pkg)));
    }
}
