#include "rpmcontroller.h"
#include "constants.h"
#include "package.h"
#include "rpmbuttonwidget.h"
#include "rpmcontrollerwidget.h"
#include "rpmfilesdialog.h"
#include "rpminfodialog.h"

#include <QDebug>
#include <QFileDialog>
#include <QRegularExpression>
#include <QRegularExpressionMatch>

class RpmControllerPrivate
{
public:
    RpmControllerPrivate()
        : m_currentPackage()
        , m_buttonWidget(nullptr)
        , m_controllerWidget(nullptr)
        , m_modelData()
        , m_selected(false)
    {}

    RpmControllerPrivate(const RpmControllerPrivate &) = delete;
    RpmControllerPrivate(RpmControllerPrivate &&)      = delete;

    RpmControllerPrivate &operator=(const RpmControllerPrivate &) = delete;
    RpmControllerPrivate &operator=(RpmControllerPrivate &&) = delete;

    Package m_currentPackage;

    RpmButtonWidget *m_buttonWidget;
    RpmControllerWidget *m_controllerWidget;

    std::vector<Package> m_modelData;

    bool m_selected;
};

RpmController::RpmController(DataSourceInterface *dataSource)
    : BaseController(dataSource)
    , d(new RpmControllerPrivate())
{
    updateModelData();
}

RpmController::~RpmController() {}

QString RpmController::getInterface()
{
    return RPM_INTERFACE_NAME;
}

QString RpmController::getPath()
{
    return RPM_OBJECT_PATH;
}

void RpmController::selected()
{
    d->m_buttonWidget     = new RpmButtonWidget(this);
    d->m_controllerWidget = new RpmControllerWidget(this);

    disableButtons();

    getWindow()->setButtonWidget(d->m_buttonWidget);
    getWindow()->setControllerWidget(d->m_controllerWidget);
    getWindow()->selectInterface(getInterface());

    d->m_controllerWidget->setPackagesList(d->m_modelData);

    d->m_controllerWidget->setFilter(QString());
    getWindow()->setSearchText(QString());

    d->m_selected = true;
}

void RpmController::deSelected()
{
    d->m_currentPackage.clear();

    d->m_selected = false;
}

void RpmController::onWindowClose() {}

void RpmController::update()
{
    deSelected();
    updateModelData();
    selected();
}

void RpmController::onFilterTextChanged(const QString text)
{
    d->m_controllerWidget->setFilter(text);
}

void RpmController::onRemoveButtonPressed()
{
    if (!d->m_selected)
    {
        return;
    }

    if (d->m_currentPackage.m_name.isEmpty())
    {
        return;
    }

    getDataSource()->rpmRemove(d->m_currentPackage.m_name);
}

void RpmController::onInstallButtonPressed()
{
    auto rpmFiles = QFileDialog::getOpenFileNames(nullptr,
                                                  QObject::tr("RPM file"),
                                                  "~/",
                                                  QObject::tr("RPM Files (*.rpm)"));

    if (rpmFiles.isEmpty())
    {
        return;
    }

    getDataSource()->rpmInstall(rpmFiles.join(" "));
}

void RpmController::onInfoButtonPressed()
{
    if (!d->m_selected || d->m_currentPackage.m_name.isEmpty())
    {
        return;
    }

    RpmInfoDialog dialog;

    QString info;
    auto backendResult = getDataSource()->rpmInfo(d->m_currentPackage.m_name);
    if (backendResult.has_value())
    {
        info = backendResult.value();
    }
    else
    {
        qWarning() << "Failed to get package info from RPM";
    }

    dialog.setInfoText(info);

    dialog.exec();
}

void RpmController::onFilesButtonPressed()
{
    if (!d->m_selected || d->m_currentPackage.m_name.isEmpty())
    {
        return;
    }

    RpmFilesDialog dialog;

    QStringList files;

    auto backendResult = getDataSource()->rpmFiles(d->m_currentPackage.m_name);
    if (backendResult.has_value())
    {
        files = backendResult.value();
    }
    else
    {
        qWarning() << "Failed to get package list from RPM";
    }

    dialog.addFiles(files);

    dialog.exec();
}

void RpmController::onItemClicked(const Package &pkg)
{
    d->m_currentPackage = pkg;
    enableButtons();
}

void RpmController::disableButtons()
{
    d->m_buttonWidget->disableButtons();
}

void RpmController::enableButtons()
{
    d->m_buttonWidget->enableButtons();
}

void RpmController::updateModelData()
{
    QStringList packages;
    auto listResult = getDataSource()->rpmList();
    if (listResult.has_value())
    {
        packages = listResult.value();
    }
    else
    {
        qWarning() << "Failed to get package list from RPM";
    }

    QRegularExpression regExp(
        "^(.+)-((?:\\d+(?:\\.\\w+)*|\\w*))-(alt\\d+.*).(x86_64|ppc64le|i586|armh|aarch64|noarch)$");
    for (auto &pkg : packages)
    {
        auto match = regExp.match(pkg);
        if (match.hasMatch())
        {
            d->m_modelData.push_back(
                Package(match.captured(1), match.captured(2), match.captured(3), match.captured(4)));
        }
        else
        {
            d->m_modelData.push_back(Package(pkg, "", "", ""));
        }
    }
}
