#include "datasource/dbusdatasource.h"
#include "constants.h"

#include <optional>
#include <QDBusConnection>
#include <QDBusInterface>
#include <QDBusMessage>
#include <QDBusReply>
#include <QDebug>
#include <QStringList>

DBusDataSource::DBusDataSource(QString serviceName, QString path, QString ifaceName, QString methodName)
    : m_connection(QDBusConnection::systemBus())
    , m_serviceName(serviceName)
    , m_alteratorPath(path)
    , m_alteratorIfaceName(ifaceName)
    , m_alteratorMethodName(methodName)
{
    QStringList rpmObjects = objectByPath(RPM_INTERFACE_NAME);

    if (!rpmObjects.isEmpty())
    {
        m_rpmPath = rpmObjects.at(0);
    }

    QStringList aptObjects = objectByPath(APT_INTERFACE_NAME);

    if (!aptObjects.isEmpty())
    {
        m_aptPath = aptObjects.at(0);
    }

    QStringList repoObjects = objectByPath(REPO_INTERFACE_NAME);

    if (!repoObjects.isEmpty())
    {
        m_repoPath = repoObjects.at(0);
    }
}

DBusDataSource::~DBusDataSource() {}

QStringList DBusDataSource::getObjectsPathsByIface(QString iface)
{
    return objectByPath(iface);
}

std::optional<QStringList> DBusDataSource::rpmList()
{
    if (m_rpmPath.isEmpty())
    {
        qWarning() << "Path to RPM object is empty, RPM functionality is unavailable";
        return std::nullopt;
    }

    auto result = dbusCall<QStringList>(m_rpmPath, RPM_INTERFACE_NAME, RPM_LIST_METHOD_NAME);
    if (!result.has_value())
    {
        return std::nullopt;
    }

    auto packages = QStringList();
    for (QString &package : result.value())
    {
        packages.append(package);
    }
    return packages;
}

std::optional<int> DBusDataSource::rpmInstall(const QString package)
{
    if (m_rpmPath.isEmpty())
    {
        qWarning() << "Path to RPM object is empty, RPM functionality is unavailable";
        return std::nullopt;
    }

    return dbusCall<int>(m_rpmPath, RPM_INTERFACE_NAME, RPM_INSTALL_METHOD_NAME, package);
}

std::optional<int> DBusDataSource::rpmRemove(const QString package)
{
    if (m_rpmPath.isEmpty())
    {
        qWarning() << "Path to RPM object is empty, RPM functionality is unavailable";
        return std::nullopt;
    }

    return dbusCall<int>(m_rpmPath, RPM_INTERFACE_NAME, RPM_REMOVE_METHOD_NAME, package);
}

std::optional<QString> DBusDataSource::rpmInfo(const QString package)
{
    if (m_rpmPath.isEmpty())
    {
        qWarning() << "Path to RPM object is empty, RPM functionality is unavailable";
        return std::nullopt;
    }

    auto result = dbusCall<QStringList>(m_rpmPath, RPM_INTERFACE_NAME, RPM_INFO_METHOD_NAME, package);
    if (!result.has_value())
    {
        return std::nullopt;
    }

    auto info = QString("");
    for (QString &line : result.value())
    {
        info.append(line + "\n");
    }

    return info;
}

std::optional<QStringList> DBusDataSource::rpmFiles(const QString package)
{
    if (m_rpmPath.isEmpty())
    {
        qWarning() << "Path to RPM object is empty, RPM functionality is unavailable";
        return std::nullopt;
    }

    return dbusCall<QStringList>(m_rpmPath, RPM_INTERFACE_NAME, RPM_FILES_METHOD_NAME, package);
}

std::optional<int> DBusDataSource::aptInstall(const QString package)
{
    if (m_aptPath.isEmpty())
    {
        qWarning() << "Path to APT object is empty, APT functionality is unavailable";
        return std::nullopt;
    }

    return dbusCall<int>(m_aptPath, APT_INTERFACE_NAME, APT_INSTALL_METHOD_NAME, package);
}

std::optional<int> DBusDataSource::aptReinstall(const QString package)
{
    if (m_aptPath.isEmpty())
    {
        qWarning() << "Path to APT object is empty, APT functionality is unavailable";
        return std::nullopt;
    }

    return dbusCall<int>(m_aptPath, APT_INTERFACE_NAME, APT_REINSTALL_METHOD_NAME, package);
}

std::optional<int> DBusDataSource::aptRemove(const QString package)
{
    if (m_aptPath.isEmpty())
    {
        qWarning() << "Path to APT object is empty, APT functionality is unavailable";
        return std::nullopt;
    }

    return dbusCall<int>(m_aptPath, APT_INTERFACE_NAME, APT_REMOVE_METHOD_NAME, package);
}

std::optional<int> DBusDataSource::aptUpdate()
{
    if (m_aptPath.isEmpty())
    {
        qWarning() << "Path to APT object is empty, APT functionality is unavailable";
        return std::nullopt;
    }

    QDBusInterface iface(m_serviceName, m_aptPath, APT_INTERFACE_NAME, m_connection);

    if (!iface.isValid())
    {
        qWarning() << "APT interface on DBus is invalid, APT functionality is unavailable";
        return std::nullopt;
    }

    QDBusPendingCall call = iface.asyncCall(APT_UPDATE_METHOD_NAME);
    auto updateWatcher    = new QDBusPendingCallWatcher(call, this);
    QObject::connect(updateWatcher, &QDBusPendingCallWatcher::finished, this, &DBusDataSource::aptUpdateSlot);

    return std::optional<int>{0};
}

void DBusDataSource::aptUpdateSlot(QDBusPendingCallWatcher *call)
{
    emit(aptUpdateFinished(call));
}

std::optional<std::vector<QString>> DBusDataSource::aptListAllPackages()
{
    if (m_rpmPath.isEmpty())
    {
        qWarning() << "Path to RPM object is empty, RPM functionality is unavailable";
        return std::nullopt;
    }

    auto result = dbusCall<QStringList>(m_aptPath, APT_INTERFACE_NAME, APT_LIST_ALL_PACKAGES_METHOD_NAME);
    if (!result.has_value())
    {
        return std::nullopt;
    }

    auto packages = std::vector<QString>();
    for (QString &package : result.value())
    {
        packages.push_back(package);
    }
    return std::optional<std::vector<QString>>{packages};
}

std::optional<QString> DBusDataSource::aptLastUpdate()
{
    if (m_rpmPath.isEmpty())
    {
        qWarning() << "Path to RPM object is empty, RPM functionality is unavailable";
        return std::nullopt;
    }

    auto result = dbusCall<QStringList>(m_aptPath, APT_INTERFACE_NAME, APT_LAST_UPDATE_METHOD_NAME);
    if (!result.has_value())
    {
        return std::nullopt;
    }

    auto lastUpdate = result.value()[0];
    return std::optional<QString>{lastUpdate};
}

QStringList DBusDataSource::objectByPath(QString iface)
{
    QStringList result;

    QDBusInterface interface(m_serviceName, m_alteratorPath, m_alteratorIfaceName, m_connection);

    if (!interface.isValid())
    {
        return result;
    }

    const QDBusReply<QList<QDBusObjectPath>> reply = interface.call(m_alteratorMethodName, iface);

    if (!reply.isValid())
    {
        return result;
    }

    for (const QDBusObjectPath &path : reply.value())
    {
        result.append(path.path());
    }

    return result;
}

std::optional<int> DBusDataSource::repoAdd(const QString repo)
{
    if (m_repoPath.isEmpty())
    {
        qWarning() << "Path to RPM_repo object is empty, RPM_repo functionality is unavailable";
        return std::nullopt;
    }

    return dbusCall<int>(m_repoPath, REPO_INTERFACE_NAME, REPO_ADD_METHOD_NAME, repo);
}

std::optional<QStringList> DBusDataSource::repoList()
{
    if (m_repoPath.isEmpty())
    {
        qWarning() << "Path to RPM_repo object is empty, RPM_repo functionality is unavailable";
        return std::nullopt;
    }

    return dbusCall<QStringList>(m_repoPath, REPO_INTERFACE_NAME, REPO_LIST_METHOD_NAME);
}

std::optional<int> DBusDataSource::repoRemove(const QString repo)
{
    if (m_repoPath.isEmpty())
    {
        qWarning() << "Path to RPM_repo object is empty, RPM_repo functionality is unavailable";
        return std::nullopt;
    }

    return dbusCall<int>(m_repoPath, REPO_INTERFACE_NAME, REPO_REMOVE_METHOD_NAME, repo);
}

template<typename T, std::enable_if_t<std::is_same<T, int>::value || std::is_same<T, QStringList>::value, bool>>
std::optional<T> DBusDataSource::dbusCall(const QString &path,
                                          const QString &interface,
                                          const QString &method,
                                          const QString &arg)
{
    if (path.isEmpty())
    {
        qWarning() << "Path to object is empty";
        return std::nullopt;
    }

    QDBusInterface iface(m_serviceName, path, interface, m_connection);

    if (!iface.isValid())
    {
        qWarning() << QString("Interface \"%1\" on DBus is invalid, corresponding functionality is unavailable")
                          .arg(interface);
        return std::nullopt;
    }

    QDBusMessage reply;
    if (!arg.isNull())
    {
        reply = iface.call(method, arg);
    }
    else
    {
        reply = iface.call(method);
    }

    if (reply.type() == QDBusMessage::ErrorMessage)
    {
        qWarning() << QString("Method \"%1\" failed with error message \"%2\"").arg(method).arg(reply.errorMessage());
        return std::nullopt;
    }

    if (reply.type() != QDBusMessage::QDBusMessage::ReplyMessage)
    {
        qWarning() << QString("Reply from method \"%1\" is invalid").arg(method);
        return std::nullopt;
    }

    bool isValid = [&]() {
        if constexpr (std::is_same<T, QStringList>::value)
        {
            return reply.arguments().count() != 2 && !reply.arguments()[0].canConvert<QStringList>()
                   && !reply.arguments()[1].canConvert<QStringList>() && !reply.arguments()[2].canConvert<int>();
        }
        else if constexpr (std::is_same<T, int>::value)
        {
            return reply.arguments().count() != 2 && !reply.arguments()[0].canConvert<QStringList>()
                   && !reply.arguments()[1].canConvert<int>();
        }
    }();

    if (isValid)
    {
        qWarning() << QString("Reply from method \"%1\" does not comply with expected API").arg(method);
        return std::nullopt;
    }

    QStringList errorMessages;
    int exitCode;
    auto result = [&]() -> std::optional<T> {
        if constexpr (std::is_same<T, QStringList>::value)
        {
            errorMessages = reply.arguments()[1].toStringList();
            exitCode      = reply.arguments()[2].toInt();
            return exitCode != 0 ? std::nullopt : std::make_optional(reply.arguments()[0].toStringList());
        }
        else if constexpr (std::is_same<T, int>::value)
        {
            errorMessages = reply.arguments()[0].toStringList();
            exitCode      = reply.arguments()[1].toInt();
            return exitCode;
        }
    }();

    if (exitCode != 0)
    {
        qWarning() << QString("Method \"%1\" failed with exit code \"%2\" and error message: \"%3\"")
                          .arg(method)
                          .arg(exitCode)
                          .arg(errorMessages.join("\n"));
    }

    return result;
}
