#include "mainwindow.h"
#include "mainwindowsettings.h"
#include "ui_mainwindow.h"

#include <QShortcut>
#include <QStandardItem>

class MainWindowPrivate
{
public:
    MainWindowPrivate(Ui::MainWindow *ui)
        : m_ui(ui)
        , m_controller(nullptr)
        , quitShortcut(nullptr)
        , settings(nullptr)
    {}

    ~MainWindowPrivate() = default;

    MainWindowPrivate(const MainWindowPrivate &) = delete;
    MainWindowPrivate(MainWindowPrivate &&)      = delete;

    MainWindowPrivate &operator=(const MainWindowPrivate &) = delete;
    MainWindowPrivate &operator=(MainWindowPrivate &&) = delete;

    std::unique_ptr<Ui::MainWindow> m_ui{nullptr};
    ControllerInterface *m_controller{nullptr};
    QAbstractItemModel *m_model{nullptr};

    QShortcut *quitShortcut{nullptr};
    std::unique_ptr<MainWindowSettings> settings{nullptr};
};

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent)
    , d(new MainWindowPrivate(new Ui::MainWindow))
{
    d->m_ui->setupUi(this);

    d->quitShortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_Q), this);

    d->settings = std::make_unique<MainWindowSettings>(this, d->m_ui.get());
    d->settings->restoreSettings();

    connect(d->m_ui->centralwidget, &PackagesWidget::filterTextChanged, this, &MainWindow::onFilterTextChanged);
    connect(d->quitShortcut, &QShortcut::activated, this, &MainWindow::closeFromButtonWidget);
}

MainWindow::~MainWindow() {}

void MainWindow::setController(ControllerInterface *controller)
{
    if (controller)
    {
        d->m_controller = controller;
    }
}

ControllerInterface *MainWindow::getController()
{
    return d->m_controller;
}

void MainWindow::setModel(QAbstractItemModel *model)
{
    if (!model)
    {
        return;
    }
    d->m_ui->centralwidget->setModel(model);

    d->m_model = model;

    connect(d->m_ui->centralwidget, &PackagesWidget::selectionChanged, this, &MainWindow::onManagerChanged);
}

void MainWindow::selectInterface(QString iface)
{
    if (iface.isEmpty())
    {
        return;
    }

    QString name = iface.mid(iface.lastIndexOf(QString(".")) + 1, iface.size() - iface.lastIndexOf(QString(".")) - 1);

    name.chop(1);

    QModelIndexList indexes = d->m_model->match(d->m_model->index(0, 0), Qt::DisplayRole, name);

    d->m_ui->centralwidget->setIndex(indexes.at(0));

    int i = 0;
    i++;
}

void MainWindow::setButtonWidget(QWidget *widget)
{
    d->m_ui->centralwidget->setButtonWidget(widget);
}

void MainWindow::setControllerWidget(QWidget *widget)
{
    d->m_ui->centralwidget->setControllerWidget(widget);
}

void MainWindow::closeFromButtonWidget()
{
    this->close();
}

void MainWindow::closeEvent(QCloseEvent *event)
{
    d->settings->saveSettings();
    d->m_controller->onWindowClose();
}

void MainWindow::onManagerChanged(const QItemSelection &newSelection, const QItemSelection &previousSelection)
{
    emit managerChanged(newSelection, previousSelection);
}

void MainWindow::onFilterTextChanged(const QString text)
{
    d->m_controller->onFilterTextChanged(text);
}

void MainWindow::setSearchText(const QString text)
{
    d->m_ui->centralwidget->setSearchText(text);
}
